/**
 * @package Openspeleo datalogger
 * @author Alessandro Vernassa <speleoalex@gmail.com>
 * @copyright Copyright (c) 2012
 * @license http://opensource.org/licenses/gpl-license.php GNU General Public License
 */
/*
Il Datalogger utilizza 6 pin. Analog 4 e 5 sono per I2C. 
La scheda SD utilizza i pin digitali 13, 12, 11, e 10. 
I primi tre sono praticamente obbligatori. 
Se davvero hai bisogno del pin 10, e' possibile modificare il 
file di intestazione della libreria e cambiarla dal pin 10 al pin 
qualsiasi altro. Ma e' necessario avere il pin 10 come uscita, 
se e' usato come un ingresso, l'interfaccia SD non funzionera'. 
I sensori sono collegati su A0,A1,2,3
Led rosso:8
Led verde:9
0=seriale rx
1=seriale tx

Riassunto pin utilizzati senza lcd:
A0(so),A1(s1),A4(i2c),A5(i2c),
0(srx),1(stx),2(s0),3(s1),8(lL1),9(L2),10(sd),11(sd),12(sd),13(sd)
Riassunto pin utilizzati con lcd:
A0(so),A1(s1),A4(i2c),A5(i2c),
0(srx),1(stx),2(s0),3(s1),4(lcd),5(lcd),6(lcd),7(lcd),8(lcd),9(lcd),10(sd),11(sd),12(sd),13(sd)
collegare lcd ad ingressi 7, 6, 5, 4, 9, 8
 */


#include <Wire.h>
#include "RTClib.h"
#include <SD.h>

#define LCD 0

#if LCD
#include <LiquidCrystal.h>
//con l'lcd occorre scollegare i led verde e rosso e utilizzare i due ingressi per l'lcd
LiquidCrystal lcd(7, 6, 5, 4, 9, 8);
static bool lcdPresent = LCD;
#endif

#define S0 A0					//sensore ldr
#define S1 A1					//sensore di temperatura
#define S0_S 2					//alimentazione S0
#define S1_S 3					//alimentazione S1

#define EXTERNAL_VOLTAGE_REF 0  		//mettere a 1 se 3.2 e' collegato ad aref
#define INTERNAL_VOLTAGE_REF 0			//set 1 for internal voltage
#define redLEDpin 8				//pin red led
#define greenLEDpin 9				//pin green led
#define MAX_ROWS_PER_FILE 10000			//max rows per file
#define CONF_FILE "CONFIG.INI"  		//configuration file
#define LOG_INTERVAL_ms 10000     		//log interval in mmilliseconds
#define SENSORPOWERANTICIPATION_ms 5000         //sensor power anticipation in ms
#define CHIP_SD 10				//pin sd
#define SYNC_INTERVAL_ms 10000			//mills between calls to flush() - to write data to the card
#define BANDGAP_VOLTAGE 1.1			//internal voltage
#if EXTERNAL_VOLTAGE_REF
#define AREF_VOLTAGE 3.3			//external voltage
#elif INTERNAL_VOLTAGE_REF
#define AREF_VOLTAGE BANDGAP_VOLTAGE            //internal voltage
static float bandgap_voltage = BANDGAP_VOLTAGE;
#else
#define AREF_VOLTAGE 5.0			//default voltage
#endif
#define NUM_READS 100


//--------------macro---------------------->
#define LOGSTRING(str) if (logfileOpened){logfile.print(str);dataToWrite++;}if(echoToSerial){Serial.print(str);}//if(lcdPresent){lcd.print(str);}
#define LOGSTRINGLN(str) if (logfileOpened){logfile.println(str);dataToWrite++;}if(echoToSerial){Serial.println(str);}//if(lcdPresent){lcd.println(str);}
//--------------macro----------------------<

//-------------globals--------------------->
static File logfile;
static File settingFile;
static bool sdPresent = false;
static bool rtcPresent = false;
static bool logfileOpened = false;
static bool echoToSerial = true;
static bool PowerSensorIsHIGHT = false;
static unsigned int log_interval_ms = LOG_INTERVAL_ms; //log interval
static unsigned int SensorPowerAnticipation_ms = SENSORPOWERANTICIPATION_ms;
static unsigned long TimeCurrent = 0;
static unsigned long TimeTarget = 0;
static unsigned long TimeTargetFileWrite = 0;
static unsigned long TimeTargetPower = 0;
static unsigned long LogCounter = 0; //counter
static unsigned long dataToWrite = 0;
static RTC_DS1307 RTC; // define the Real Time Clock object
static float aref_voltage = AREF_VOLTAGE; // we tie 3.3V to ARef and measure it with a multimeter!

char strDate[] = "0000-00-00 00:00:00";
//-------------globals---------------------<

/**
 * convert adc value to temperature in celsius
 */
float DL_ValToTemp(float val) {
    return (aref_voltage * val * 100.0) / 1024;
}

/**
 *
 */
long DL_readVcc() {
    // Read 1.1V reference against AVcc
    // set the reference to Vcc and the measurement to the internal 1.1V reference
#if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
    ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
#elif defined (__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
    ADMUX = _BV(MUX5) | _BV(MUX0);
#elif defined (__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
    ADMUX = _BV(MUX3) | _BV(MUX2);
#else
    ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
#endif
    delay(2); // Wait for Vref to settle
    ADCSRA |= _BV(ADSC); // Start conversion
    while (bit_is_set(ADCSRA, ADSC))
        ; // measuring
    uint8_t low = ADCL; // must read ADCL first - it then locks ADCH
    uint8_t high = ADCH; // unlocks both
    long result = (high << 8) | low;
    result = 1125300L / result; // Calculate Vcc (in mV); 1125300 = 1.1*1023*1000
    return result; // Vcc in millivolts
}

/**
 * 
 */
static void DL_fixStrFilename(char * str) {
    for (unsigned int i = 0; i < strlen(str) && i < 7; i++) {
        if (str[i] == ' ') {
            str[i] = '0';
        }
    }
}

/**
 *
 */
static char * DL_strNow(void) {
    DateTime  now;
    now = RTC.now();
    sprintf(strDate, "%4d-%2d-%2d_%2d:%2d:%2d", now.year(), now.month(),
            now.day(), now.hour(), now.minute(), now.second());
    for (unsigned int i = 0; i < strlen(strDate) && i < 18; i++) {
        if (strDate[i] == ' ') {
            strDate[i] = '0';
        } else {
            if (strDate[i] == '_') {
                strDate[i] = ' ';
            }
        }
    }
    return strDate;
}

/**
 *
 */
static void DL_openLogFile() {
    if (sdPresent == true) {
        if (logfileOpened) {
            logfile.flush();
            logfile.close();
            logfileOpened = false;
        }
        char filename[] = "LOG0000.CSV";
        for (uint8_t i = 0; i < 9999; i++) {
            sprintf(filename, "LOG%4d.CSV", i);
            DL_fixStrFilename(filename);
            if (!SD.exists(filename)) {
                // only open a new file if it doesn't exist
                logfile = SD.open(filename, FILE_WRITE);
                if (logfile) {
                    logfileOpened = true;
                } else {
                    logfileOpened = false;
                }
                break;
            }
        }
        if (!logfileOpened) {
            Serial.print("Couldnt create file:");
            Serial.println(filename);
        } else {
            Serial.print("Log to:");
            Serial.println(filename);
        }
    } else {
        Serial.println("Log:serial");
    }
    LOGSTRINGLN("\"count\"\t\"date\"\t\"light\"\t\"temperature\"\t\"S0V\"\t\"S1V\"\t\"VCC\"");
}

/**
 * 
 */
static bool CONF_is_valid_char(char character) {
    if (isalnum(character) || character == '_') {
        return true;
    }
    return false;
}

/**
 * 
 */
static int CONF_getConfValueInt(char* filename, char* key, int defaultValue = 0) {
    File myFile;
    char character;
    String description = "";
    String value = "";
    boolean valid = true;
    int ret = defaultValue;
    if (!SD.exists(filename)) {
        //Serial.println("file not exists");
        return defaultValue;
    }
    myFile = SD.open(filename);

    while (myFile.available()) {
        description = "";
        character = myFile.read();

        //cerca una linea valida----->
        if (!CONF_is_valid_char(character)) {
            // Comment - ignore this line
            while (character != '\n' && myFile.available()) {
                character = myFile.read();
            };
            continue;
        }
        //cerca una linea valida-----<
        //----riempo la descrizione---->
        do {
            description.concat(character);
            character = myFile.read();
            if (!myFile.available()) {
                myFile.close();
                return defaultValue;
            }
        } while (CONF_is_valid_char(character));
        //----riempo la descrizione----<
        //-------elimino gli spazi------->
        if (character == ' ') {
            do {
                character = myFile.read();
                if (!myFile.available()) {
                    myFile.close();
                    return defaultValue;
                }
            } while (character == ' ');
        }
        //-------elimino gli spazi-------<
        if (character == '=') {
            if (description == key) {
                //riempo il valore
                //Serial.println("trovato... scorro:");
                //-------elimino gli spazi------->
                do {
                    character = myFile.read();
                    if (!myFile.available()) {
                        myFile.close();
                        return defaultValue;
                    }
                } while (character == ' ');
                //-------elimino gli spazi-------<
                value = "";
                valid = true;
                while (character != '\n' && character != '\r') {
                    //Serial.println(character);
                    if (isdigit(character)) {
                        value.concat(character);
                    } else if (character != '\n' && character != '\r') {
                        // Use of invalid values
                        valid = false;
                    }
                    character = myFile.read();
                };
                if (valid) {
                    // Convert string to array of chars
                    char charBuf[value.length() + 1];
                    value.toCharArray(charBuf, value.length() + 1);
                    // Convert chars to integer
                    ret = atoi(charBuf);
                    myFile.close();
                    return ret;
                }
            }
        } else {
            while (character != '\n' && myFile.available()) {
                character = myFile.read();
            };
            continue;
        }
    }

    return ret;
}

/**
 * Fa N letture, prende i 10 centrali e fa la media di quelli
 */
static float DL_analogReadAndFilter(uint8_t sensorpin) {
    // read multiple values and sort them to take the mode
    analogRead(sensorpin);
    delay(10);
    int sortedValues[NUM_READS];
    for (int i = 0; i < NUM_READS; i++) {
        int value = analogRead(sensorpin);
        int j;
        if (value < sortedValues[0] || i == 0) {
            j = 0; //insert at first position
        } else {
            for (j = 1; j < i; j++) {
                if (sortedValues[j - 1] <= value && sortedValues[j] >= value) {
                    // j is insert position
                    break;
                }
            }
        }
        for (int k = i; k > j; k--) {
            // move all values higher than current reading up one position
            sortedValues[k] = sortedValues[k - 1];
        }
        sortedValues[j] = value; //insert current reading
        delay(1);
    }
    //return scaled mode of 10 values
    float returnval = 0;
    for (int i = NUM_READS / 2 - 5; i < (NUM_READS / 2 + 5); i++) {
        returnval += sortedValues[i];
    }
    returnval = returnval / 10;
#if INTERNAL_VOLTAGE_REF
    return returnval * (bandgap_voltage * 1000) / 1023;
#else
    return returnval;
#endif
}

/**
 *
 */
static void DL_initConf() {
    if (sdPresent) {
        if (!SD.exists(CONF_FILE)) {
            Serial.print(CONF_FILE);
            Serial.println(F(" not exist"));
            settingFile = SD.open(CONF_FILE, FILE_WRITE);
            if (settingFile) {
                //create new file config.ini------>
                Serial.print(F("new file"));
                Serial.println(CONF_FILE);
                settingFile.print(F("log_interval_ms="));
                settingFile.println(LOG_INTERVAL_ms);
                settingFile.print(F("anticipation_ms="));
                settingFile.println(SENSORPOWERANTICIPATION_ms);
                settingFile.flush();
                settingFile.close();
                //create new file config.ini------<
            } else {
                Serial.print(F("error to create "));
                Serial.println(CONF_FILE);
            }
        } else {
            Serial.print(CONF_FILE);
            Serial.println(" already exist");
        }
        Serial.flush();
        //read from CONFIG.INI------->
        log_interval_ms = CONF_getConfValueInt(CONF_FILE, "log_interval_ms", LOG_INTERVAL_ms);
        SensorPowerAnticipation_ms = CONF_getConfValueInt(CONF_FILE, "anticipation_ms", SENSORPOWERANTICIPATION_ms);
        //read from CONFIG.INI-------<
        Serial.print(F("Interval(ms)="));
        Serial.println(log_interval_ms);
        Serial.print(F("Anticipation(ms)="));
        Serial.println(SensorPowerAnticipation_ms);
        Serial.flush();
    }
    if (log_interval_ms == 0) {
        log_interval_ms = LOG_INTERVAL_ms;
        SensorPowerAnticipation_ms = SENSORPOWERANTICIPATION_ms;
        Serial.print(F("Force to "));
        Serial.println(LOG_INTERVAL_ms);
    }
}

/**
 * The setup function is called once at startup of the sketch
 */
void setup() {
    int n = 0;

    //--init serial------------------------------>
    Serial.begin(9600);
    //--init serial------------------------------<
    //int unixtime = DateTime(__DATE__, __TIME__).unixtime();
    sdPresent = false;
    rtcPresent = false;
#if LCD
    if (lcdPresent) {
        lcd.begin(16, 2);
        lcd.print("Openspeleo.org");
    }
#endif
    logfileOpened = false;
    PowerSensorIsHIGHT = false;
    pinMode(S0_S, OUTPUT);
    pinMode(S1_S, OUTPUT);
    // initialize the SD card ------------------->
    //Serial.println("Initializing SD");
    pinMode(CHIP_SD, OUTPUT);
    Serial.print(F("SD initialization "));
    if (!SD.begin(CHIP_SD)) {
        sdPresent = false;
        Serial.println(F("failed"));
    } else {
        Serial.println(F("successful"));
        sdPresent = true;
    }
    Serial.flush();
    DL_initConf();
    // initialize the SD card -------------------<
    // connect to RTC --------------------------->
    Wire.begin();
    unsigned long unixtime = DateTime(__DATE__, __TIME__).unixtime();
    if (RTC.begin()) {
        rtcPresent = true;
        if (!RTC.isrunning()) {
            Serial.println(F("Clock not running"));
            RTC.adjust(unixtime);
        }
        Serial.print(F("Current clock:"));
        Serial.println(DL_strNow());
        Serial.println(F("Type any character to synchronize the clock with the PC time:"));
        while (!Serial.available()) {
            //Serial.print("Internal clock:");
            Serial.print(10 - n);
            Serial.print(' ');
            if (n >= 10) {
                break;
            }
            delay(1000);
            n++;
        }
        if (n < 10) {
            // "Clock adjust";
            Serial.println();
            Serial.println(F("Clock has been synchronized:"));
            Serial.println(DL_strNow());
            RTC.adjust(unixtime + n + 15); //sulla mia seriale è sempre 15 secondi indietro
        } else {
            Serial.println();
        }

        //Serial.print("Internal clock:");
        
    } else {
        rtcPresent = false;
        Serial.println(F("RTC Failed"));
    }

    // connect to RTC ---------------------------<
    delay(10);
    TimeCurrent = millis();
    TimeTarget = TimeCurrent + log_interval_ms;
    TimeTargetFileWrite = TimeCurrent + SYNC_INTERVAL_ms;
    //-----------init analog input----------------------------------------->
#if EXTERNAL_VOLTAGE_REF
    analogReference(EXTERNAL); //aref esterna da 3.2
#elif INTERNAL_VOLTAGE_REF
    analogReference(INTERNAL); //aref esterna da 5.0
#else
    analogReference(DEFAULT); //aref esterna da 5.0
#endif
    //turn ON sensors ------->
    digitalWrite(S0_S, LOW);
    digitalWrite(S1_S, LOW);
    PowerSensorIsHIGHT = false;
    //turn ON sensors -------<
    //---faccio una lettura a vuoto per scaricare il condensatore -------->
    for (int i = 0; i < 10; i++) {
        DL_analogReadAndFilter(S0);
        DL_analogReadAndFilter(S1);
        delay(20);
    }
    //---faccio una lettura a vuoto per scaricare il condensatore --------<
    //-----------init analog input-----------------------------------------<
    TimeTarget = millis();
    TimeTargetPower = TimeTarget - SensorPowerAnticipation_ms;
    //DL_openLogFile();


}

/**
 * Loop principale.
 */
void loop() {
    float photocellReading = 0.0;
    float tempReading = 0.0;
    float s0v = 0.0;
    float s1v = 0.0;
    long vcc = 0;
    TimeCurrent = millis();
    // turn ON the sensor before the measure -------------------------------------->
    if (PowerSensorIsHIGHT == false && TimeCurrent >= TimeTargetPower) {
        digitalWrite(S0_S, HIGH);
        digitalWrite(S1_S, HIGH);
        // Serial.println("ON");
        PowerSensorIsHIGHT = true;
    }
    // turn ON the sensor before the measure --------------------------------------<
    if (TimeCurrent >= TimeTarget) {
        TimeTarget = TimeCurrent + log_interval_ms;
        TimeTargetPower = TimeTarget - SensorPowerAnticipation_ms;

        if (LogCounter == 0 || (LogCounter >= MAX_ROWS_PER_FILE
                && (LogCounter % MAX_ROWS_PER_FILE == 0))) {
            DL_openLogFile();
        }
        //-----------sensors------------------------------------------------------->
        photocellReading = DL_analogReadAndFilter(S0);
        tempReading = DL_ValToTemp(DL_analogReadAndFilter(S1));
        s0v = (photocellReading / 1023.0) * AREF_VOLTAGE;
        s1v = (tempReading / 1023.0) * AREF_VOLTAGE;
        vcc = DL_readVcc();
        //-----------sensors-------------------------------------------------------<
        LogCounter++;
        LOGSTRING(LogCounter);
        LOGSTRING("\t");
        LOGSTRING("\"");
        LOGSTRING(DL_strNow());
        LOGSTRING("\"");
        LOGSTRING("\t");
        LOGSTRING(photocellReading);
        LOGSTRING("\t");
        LOGSTRING(tempReading);
        LOGSTRING("\t");
        LOGSTRING(s0v);
        LOGSTRING("\t");
        LOGSTRING(s1v);
        LOGSTRING("\t");
        LOGSTRINGLN(vcc);
        //LOGSTRINGLN("");
        if (PowerSensorIsHIGHT == true && TimeCurrent < TimeTargetPower) {
            // Serial.println("OFF");
            digitalWrite(S0_S, LOW);
            digitalWrite(S1_S, LOW);
            PowerSensorIsHIGHT = false;
        }
#if LCD
        if (lcdPresent) {
            if (LogCounter % 2) {
                lcd.setCursor(0, 0);
                lcd.print("Input:");
                lcd.print(S0);
                lcd.print("    ");
                lcd.setCursor(0, 1);
                lcd.print(photocellReading);
                lcd.print("    ");
            } else {
                lcd.setCursor(0, 0);
                lcd.print("Temp:");
                lcd.print("    ");
                lcd.setCursor(0, 1);
                lcd.print(tempReading);
                lcd.print("    ");
            }
        }
#endif
    }
    if (dataToWrite != 0 && TimeCurrent >= TimeTargetFileWrite) {
        if (logfileOpened) {
            logfile.flush();
        }
        dataToWrite = 0;
        TimeTargetFileWrite = TimeCurrent + SYNC_INTERVAL_ms;
    }
}


